import asyncio
from typing import Optional

from openinference.instrumentation.langchain import (
    LangChainInstrumentor,
    get_current_span,
)
from uipath._cli._debug._bridge import ConsoleDebugBridge, UiPathDebugBridge
from uipath._cli._runtime._contracts import (
    UiPathRuntimeContext,
    UiPathRuntimeFactory,
    UiPathRuntimeResult,
)
from uipath._cli.middlewares import MiddlewareResult
from uipath._events._events import UiPathAgentStateEvent
from uipath.tracing import JsonLinesFileExporter, LlmOpsHttpExporter

from .._tracing import (
    _instrument_traceable_attributes,
)
from ._runtime._exception import LangGraphRuntimeError
from ._runtime._memory import get_memory
from ._runtime._runtime import LangGraphScriptRuntime
from ._utils._graph import LangGraphConfig


def langgraph_run_middleware(
    entrypoint: Optional[str],
    input: Optional[str],
    resume: bool,
    trace_file: Optional[str] = None,
    **kwargs,
) -> MiddlewareResult:
    """Middleware to handle LangGraph execution"""
    config = LangGraphConfig()
    if not config.exists:
        return MiddlewareResult(
            should_continue=True
        )  # Continue with normal flow if no langgraph.json

    try:
        _instrument_traceable_attributes()

        async def execute():
            context = UiPathRuntimeContext.with_defaults(**kwargs)
            context.entrypoint = entrypoint
            context.input = input
            context.resume = resume

            async with get_memory(context) as memory:
                runtime_factory = UiPathRuntimeFactory(
                    LangGraphScriptRuntime,
                    UiPathRuntimeContext,
                    runtime_generator=lambda ctx: LangGraphScriptRuntime(
                        ctx, memory, ctx.entrypoint
                    ),
                )

                runtime_factory.add_instrumentor(
                    LangChainInstrumentor, get_current_span
                )

                if trace_file:
                    runtime_factory.add_span_exporter(JsonLinesFileExporter(trace_file))

                if context.job_id:
                    runtime_factory.add_span_exporter(
                        LlmOpsHttpExporter(extra_process_spans=True)
                    )
                    await runtime_factory.execute(context)
                else:
                    debug_bridge: UiPathDebugBridge = ConsoleDebugBridge()
                    await debug_bridge.emit_execution_started("default")
                    async for event in runtime_factory.stream(context):
                        if isinstance(event, UiPathRuntimeResult):
                            await debug_bridge.emit_execution_completed(event)
                        elif isinstance(event, UiPathAgentStateEvent):
                            await debug_bridge.emit_state_update(event)

        asyncio.run(execute())

        return MiddlewareResult(
            should_continue=False,
            error_message=None,
        )

    except LangGraphRuntimeError as e:
        return MiddlewareResult(
            should_continue=False,
            error_message=e.error_info.detail,
            should_include_stacktrace=True,
        )
    except Exception as e:
        return MiddlewareResult(
            should_continue=False,
            error_message=f"Error: {str(e)}",
            should_include_stacktrace=True,
        )



import asyncio
from typing import Optional

from openinference.instrumentation.langchain import (
    LangChainInstrumentor,
    get_current_span,
)
from uipath._cli._dev._terminal import UiPathDevTerminal
from uipath._cli._runtime._contracts import UiPathRuntimeContext, UiPathRuntimeFactory
from uipath._cli._utils._console import ConsoleLogger
from uipath._cli.middlewares import MiddlewareResult

from .._tracing import _instrument_traceable_attributes
from ._runtime._memory import get_memory
from ._runtime._runtime import LangGraphScriptRuntime
from ._utils._graph import LangGraphConfig

console = ConsoleLogger()


def langgraph_dev_middleware(interface: Optional[str]) -> MiddlewareResult:
    """Middleware to launch the developer terminal"""
    config = LangGraphConfig()
    if not config.exists:
        return MiddlewareResult(
            should_continue=True
        )  # Continue with normal flow if no langgraph.json

    try:
        if interface == "terminal":
            _instrument_traceable_attributes()

            async def execute():
                context = UiPathRuntimeContext.with_defaults()

                async with get_memory(context) as memory:
                    runtime_factory = UiPathRuntimeFactory(
                        LangGraphScriptRuntime,
                        UiPathRuntimeContext,
                        lambda ctx: LangGraphScriptRuntime(ctx, memory, ctx.entrypoint),
                    )

                    runtime_factory.add_instrumentor(
                        LangChainInstrumentor, get_current_span
                    )

                    app = UiPathDevTerminal(runtime_factory)

                    await app.run_async()

            asyncio.run(execute())
        else:
            console.error(f"Unknown interface: {interface}")
    except KeyboardInterrupt:
        console.info("Debug session interrupted by user")
    except Exception as e:
        console.error(f"Error occurred: {e}")
        return MiddlewareResult(
            should_continue=False,
            should_include_stacktrace=True,
        )

    return MiddlewareResult(should_continue=False)



import os
from contextlib import asynccontextmanager

from langgraph.checkpoint.sqlite.aio import AsyncSqliteSaver
from uipath._cli._runtime._contracts import UiPathRuntimeContext


def get_connection_string(context: UiPathRuntimeContext) -> str:
    if context.runtime_dir and context.state_file:
        path = os.path.join(context.runtime_dir, context.state_file)
        if not context.resume and context.job_id is None:
            # If not resuming and no job id, delete the previous state file
            if os.path.exists(path):
                os.remove(path)
        os.makedirs(context.runtime_dir, exist_ok=True)
        return path
    return os.path.join("__uipath", "state.db")


@asynccontextmanager
async def get_memory(context: UiPathRuntimeContext):
    """Create and manage the AsyncSqliteSaver instance."""
    async with AsyncSqliteSaver.from_conn_string(
        get_connection_string(context)
    ) as memory:
        yield memory



import asyncio
from typing import Optional

from openinference.instrumentation.langchain import (
    LangChainInstrumentor,
    get_current_span,
)
from uipath._cli._debug._bridge import UiPathDebugBridge, get_debug_bridge
from uipath._cli._debug._runtime import UiPathDebugRuntime
from uipath._cli._runtime._contracts import (
    UiPathRuntimeContext,
    UiPathRuntimeFactory,
)
from uipath._cli.middlewares import MiddlewareResult
from uipath.tracing import LlmOpsHttpExporter

from .._tracing import _instrument_traceable_attributes
from ._runtime._exception import LangGraphRuntimeError
from ._runtime._memory import get_memory
from ._runtime._runtime import LangGraphScriptRuntime
from ._utils._graph import LangGraphConfig


def langgraph_debug_middleware(
    entrypoint: Optional[str], input: Optional[str], resume: bool, **kwargs
) -> MiddlewareResult:
    """Middleware to handle LangGraph execution"""
    config = LangGraphConfig()
    if not config.exists:
        return MiddlewareResult(
            should_continue=True
        )  # Continue with normal flow if no langgraph.json

    try:

        async def execute():
            context = UiPathRuntimeContext.with_defaults(**kwargs)
            context.entrypoint = entrypoint
            context.input = input
            context.resume = resume

            _instrument_traceable_attributes()

            async with get_memory(context) as memory:
                runtime_factory = UiPathRuntimeFactory(
                    LangGraphScriptRuntime,
                    UiPathRuntimeContext,
                    runtime_generator=lambda ctx: LangGraphScriptRuntime(
                        ctx, memory, ctx.entrypoint
                    ),
                    context_generator=lambda: context,
                )

                if context.job_id:
                    runtime_factory.add_span_exporter(
                        LlmOpsHttpExporter(extra_process_spans=True)
                    )

                runtime_factory.add_instrumentor(
                    LangChainInstrumentor, get_current_span
                )

                debug_bridge: UiPathDebugBridge = get_debug_bridge(context)

                async with UiPathDebugRuntime.from_debug_context(
                    factory=runtime_factory,
                    context=context,
                    debug_bridge=debug_bridge,
                ) as debug_runtime:
                    await debug_runtime.execute()

        asyncio.run(execute())

        return MiddlewareResult(
            should_continue=False,
            error_message=None,
        )

    except LangGraphRuntimeError as e:
        return MiddlewareResult(
            should_continue=False,
            error_message=e.error_info.detail,
            should_include_stacktrace=True,
        )
    except Exception as e:
        return MiddlewareResult(
            should_continue=False,
            error_message=f"Error: {str(e)}",
            should_include_stacktrace=True,
        )



import asyncio
from typing import List, Optional

from openinference.instrumentation.langchain import (
    LangChainInstrumentor,
    get_current_span,
)
from uipath._cli._evals._console_progress_reporter import ConsoleProgressReporter
from uipath._cli._evals._evaluate import evaluate
from uipath._cli._evals._progress_reporter import StudioWebProgressReporter
from uipath._cli._evals._runtime import UiPathEvalContext
from uipath._cli._runtime._contracts import (
    UiPathRuntimeContext,
    UiPathRuntimeFactory,
)
from uipath._cli._utils._eval_set import EvalHelpers
from uipath._cli.middlewares import MiddlewareResult
from uipath._events._event_bus import EventBus
from uipath.eval._helpers import auto_discover_entrypoint
from uipath.tracing import LlmOpsHttpExporter

from uipath_langchain._cli._runtime._memory import get_memory
from uipath_langchain._cli._runtime._runtime import LangGraphScriptRuntime
from uipath_langchain._cli._utils._graph import LangGraphConfig
from uipath_langchain._tracing import (
    _instrument_traceable_attributes,
)


def langgraph_eval_middleware(
    entrypoint: Optional[str], eval_set: Optional[str], eval_ids: List[str], **kwargs
) -> MiddlewareResult:
    config = LangGraphConfig()
    if not config.exists:
        return MiddlewareResult(
            should_continue=True
        )  # Continue with normal flow if no langgraph.json

    try:
        _instrument_traceable_attributes()

        async def execute():
            event_bus = EventBus()

            if kwargs.get("register_progress_reporter", False):
                progress_reporter = StudioWebProgressReporter(
                    spans_exporter=LlmOpsHttpExporter(extra_process_spans=True)
                )
                await progress_reporter.subscribe_to_eval_runtime_events(event_bus)

            console_reporter = ConsoleProgressReporter()
            await console_reporter.subscribe_to_eval_runtime_events(event_bus)

            def generate_runtime_context(
                context_entrypoint: str, **context_kwargs
            ) -> UiPathRuntimeContext:
                context = UiPathRuntimeContext.with_defaults(**context_kwargs)
                context.entrypoint = context_entrypoint
                return context

            runtime_entrypoint = entrypoint or auto_discover_entrypoint()

            eval_context = UiPathEvalContext.with_defaults(
                entrypoint=runtime_entrypoint, **kwargs
            )
            eval_context.eval_set = eval_set or EvalHelpers.auto_discover_eval_set()
            eval_context.eval_ids = eval_ids

            async with get_memory(eval_context) as memory:
                runtime_factory = UiPathRuntimeFactory(
                    LangGraphScriptRuntime,
                    UiPathRuntimeContext,
                    context_generator=lambda **context_kwargs: generate_runtime_context(
                        context_entrypoint=runtime_entrypoint,
                        **context_kwargs,
                    ),
                    runtime_generator=lambda ctx: LangGraphScriptRuntime(
                        ctx, memory, ctx.entrypoint
                    ),
                )

                if eval_context.job_id:
                    runtime_factory.add_span_exporter(
                        LlmOpsHttpExporter(extra_process_spans=True)
                    )

                runtime_factory.add_instrumentor(
                    LangChainInstrumentor, get_current_span
                )

                await evaluate(runtime_factory, eval_context, event_bus)

        asyncio.run(execute())

        return MiddlewareResult(should_continue=False)

    except Exception as e:
        return MiddlewareResult(
            should_continue=False, error_message=f"Error running evaluation: {str(e)}"
        )
